/*
** Copyright (C) 2009-2010 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.227.7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
**  A "plug-in" (currently compile-time only) for rwflowpack to
**  process IPFIX files from a directory.
**
*/


#include <silk/silk.h>

RCSIDENT("$SiLK: ipfixdirreader.c 15441 2010-03-10 15:58:28Z mthomas $");

#include "rwflowpack_priv.h"
#include <silk/skpolldir.h>
#include <silk/skdeque.h>


/* LOCAL DEFINES AND TYPEDEFS */

/* A name for this reader type. */
#define READER_TYPE_NAME  "IPFIX Directory Reader"

typedef struct pdudir_data_st {
    char             pathname[PATH_MAX];
    char            *filename;
    skPollDirQueue_t pdq;
    skPollDir_t      polldir;
    ipfixSource_t    ipfixsource;
} ipfixdir_data_t;


/* LOCAL VARIABLE DEFINITIONS */

/* The polling interval */
static uint32_t polling_interval;

/* List of ipfixdir_data_t objects saved here for shutdown purposes */
static skDeque_t ipfixdir_list = NULL;

/* LOCAL FUNCTION PROTOTYPES */

static int  readerSetup(
    fp_daemon_mode_t   *is_daemon,
    const sk_vector_t  *probe_vec,
    reader_options_t   *options);
static int  readerStart(flow_proc_t *fproc);
static int  readerStop(flow_proc_t *fproc);
static fp_get_record_result_t readerGetRecord(
    rwRec                  *out_rwrec,
    const skpc_probe_t    **out_probe,
    flow_proc_t            *fproc);
static int  readerWantProbe(skpc_probe_t *probe);
static void readerTeardown(void);

/* FUNCTION DEFINITIONS */

/*
 *  ok = readerStart(flow_processor)
 *
 *    Create a flowsource object that will read records from IPFIX
 *    files in a directory.
 *
 *    Invoked by reader_type->start_fn();
 */
static int readerStart(flow_proc_t *fproc)
{
    const char *dir;
    ipfixdir_data_t *ipfixdir_source;

    /* Should be NULL when starting */
    assert(fproc->flow_src == NULL);

    dir = skpcProbeGetPollDirectory(fproc->probe);
    assert(dir != NULL);

    ipfixdir_source = (ipfixdir_data_t *)calloc(1, sizeof(*ipfixdir_source));
    if (ipfixdir_source == NULL) {
        /* memory error.  Fail */
        ERRMSG("Out of memory error");
        return -1;
    }

    /* Set up directory polling */
    DEBUGMSG(("Creating " READER_TYPE_NAME
              " directory poller for probe %s on '%s'"),
             skpcProbeGetName(fproc->probe), dir);

    ipfixdir_source->pdq = skPollDirQueueCreate();
    if (ipfixdir_source->pdq == NULL) {
        /* memory error.  Fail */
        free(ipfixdir_source);
        ERRMSG("Could not create polldir queue");
        return -1;
    }

    ipfixdir_source->polldir = skPollDirCreate(dir, polling_interval,
                                               ipfixdir_source->pdq);
    if (ipfixdir_source->polldir == NULL) {
        skPollDirQueueDestroy(ipfixdir_source->pdq);
        free(ipfixdir_source);
        ERRMSG("Could not initiate polling on directory '%s'", dir);
        return -1;
    }

    if (skDequePushBack(ipfixdir_list, ipfixdir_source) != SKDQ_SUCCESS) {
        skPollDirQueueDestroy(ipfixdir_source->pdq);
        skPollDirDestroy(ipfixdir_source->polldir);
        free(ipfixdir_source);
        ERRMSG("Could not add ipfixdir source to queue");
        return -1;
    }

    fproc->flow_src = ipfixdir_source;

    return 0;
}


/*
 *  readerStop();
 *
 *    Stop reading records.
 *
 *    Invoked by reader_type->stop_fn();
 */
static int readerStop(flow_proc_t *fproc)
{
    ipfixdir_data_t *ipfixdir_source = (ipfixdir_data_t *)fproc->flow_src;

    if (ipfixdir_source) {
        if (ipfixdir_source->polldir) {
            DEBUGMSG("Destroying directory poller");
            skPollDirDestroy(ipfixdir_source->polldir);
        }
        if (ipfixdir_source->pdq) {
            skPollDirQueueDestroy(ipfixdir_source->pdq);
        }
    }

    return 0;
}

/*
 *  readerTeardown();
 *
 *    Destroy the IPFIX reader.
 *
 *    Invoked by reader_type->teardown_fn();
 */
static void readerTeardown(void)
{
    ipfixdir_data_t *ipfixdir_source;

    if (ipfixdir_list) {
        while (skDequePopFrontNB(ipfixdir_list, (void **)&ipfixdir_source)
               == SKDQ_SUCCESS)
        {
            if (ipfixdir_source->ipfixsource) {
                ipfixSourceDestroy(ipfixdir_source->ipfixsource);
            }
            free(ipfixdir_source);
        }
        skDequeDestroy(ipfixdir_list);
        ipfixdir_list = NULL;
    }
}


/*
 *  readerGetNextValidFile(fproc);
 *
 *    Pull the next file name from the poll-directory and create a
 *    flowsource object to read the IPFIX records in it.  Modifies the
 *    source object that 'fproc' contains.
 *
 *    Return 0 on success.  Return -1 if getting the file name fails.
 *    If unable to open the file or file not of correct form, return
 *    -2 unless the --error-dir is set, in which case move the file
 *    there and try the next file.
 */
static int readerGetNextValidFile(flow_proc_t *fproc)
{
    int rv;
    ipfixdir_data_t *ipfixdir_source = (ipfixdir_data_t *)fproc->flow_src;

    do {
        /* Get next file from the directory poller */
        rv = skPollDirGetNextFile(ipfixdir_source->pdq,
                                  ipfixdir_source->pathname,
                                  &ipfixdir_source->filename);
        if (rv != 0) {
            if ((rv == -1) &&
                (skPollDirGetError(ipfixdir_source->polldir) != PDERR_NONE))
            {
                CRITMSG("Polldir error ocurred at %s:%d",
                        __FILE__, __LINE__);
                abort();
            }
            return -1;
        }

        INFOMSG((READER_TYPE_NAME " processing '%s'"),
                ipfixdir_source->filename);

        /* Zero record count */
        fproc->rec_count_total = 0;
        fproc->rec_count_bad = 0;
        assert(ipfixdir_source->ipfixsource == NULL);

        /* open the file to create a source of records */
        ipfixdir_source->ipfixsource =
            ipfixFileSourceCreate(fproc->probe, ipfixdir_source->pathname);
        if (ipfixdir_source->ipfixsource == NULL) {
            rv = errorDirectoryInsertFile(ipfixdir_source->pathname);
            if (rv != 0) {
                /* either no --error-dir (rv == 1) or problem moving
                 * the file (rv == -1).  either way, return an error
                 * code to the caller. */
                return -2;
            }
        }
    } while (ipfixdir_source->ipfixsource == NULL);

    return 0;
}


/*
 *  readerGetRecord(&out_rwrec, &out_probe);
 *
 *    Fill 'out_rwrec' with an rw generic record from the underlying
 *    flowsource object.  If there is no flowsource object, create one
 *    by pulling a file off the queue.  Fill 'out_probe' with the
 *    probe where the flow was collected.
 *
 *    Invoked by reader_type->get_record_fn();
 */
fp_get_record_result_t readerGetRecord(
    rwRec                  *out_rwrec,
    const skpc_probe_t    **out_probe,
    flow_proc_t            *fproc)
{
    ipfixdir_data_t *ipfixdir_source = (ipfixdir_data_t *)fproc->flow_src;
    fp_get_record_result_t retVal = FP_GET_ERROR;

    /* If we don't have a source, pop a file off the valid-queue
     * and start processing it. */
    if (ipfixdir_source->ipfixsource == NULL) {
        switch (readerGetNextValidFile(fproc)) {
          case 0:
            /* Success */
            break;
          case -1:
            /* Error getting file name (maybe in shutdown?) */
            goto END;
          case -2:  /* Error opening file */
          default:  /* Unexpected value */
            retVal = FP_FATAL_ERROR;
            goto END;
        }
    }

    /* Assume we can get a record from the probe. */
    retVal = FP_RECORD;
    *out_probe = fproc->probe;

    /* Try to get a record */
    if (-1 == ipfixSourceGetGeneric(ipfixdir_source->ipfixsource, out_rwrec)) {
        /* Failed to get a record: either at EOF or got an error */

        if (0 == fproc->rec_count_total) {
            /* failed to get first record.  treat entire file as bad */
            NOTICEMSG("File '%s' does not appear to be an IPFIX file",
                      ipfixdir_source->pathname);

            if (0 == errorDirectoryInsertFile(ipfixdir_source->pathname)) {
                /* moved file to error-dir */
                retVal = FP_FILE_BREAK;
            } else {
                /* either no --error-dir or problem moving the file.
                 * either way, return an error code to the caller. */
                retVal = FP_FATAL_ERROR;
            }
        } else {
            /* we've seen good records in the past, so assume EOF. */
            retVal = FP_FILE_BREAK;

            /* Either archive the file or remove it */
            archiveDirectoryInsertOrRemove(ipfixdir_source->pathname, NULL);
        }

        /* Clear out, getting ready for next file */
        *out_probe = NULL;
        ipfixSourceDestroy(ipfixdir_source->ipfixsource);
        ipfixdir_source->ipfixsource = NULL;
    }

  END:
    return retVal;
}

/*
 * ipfixDirReaderInitialize:
 *
 *    Set initial state of this flow reader.
 *
 *    Invoked by reader_type->initialize_fn();
 */
int ipfixDirReaderInitialize(reader_type_t *reader_type)
{
    /* Set my name */
    reader_type->reader_name = READER_TYPE_NAME;

    /* Set function pointers */
    reader_type->setup_fn            = &readerSetup;
    reader_type->teardown_fn         = &readerTeardown;
    reader_type->start_fn            = &readerStart;
    reader_type->stop_fn             = &readerStop;
    reader_type->get_record_fn       = &readerGetRecord;
    reader_type->want_probe_fn       = &readerWantProbe;

    return 0;
}


/*
 *  yes_or_no = readerWantProbe(probe);
 *
 *    Return a TRUE value this reader_type is able to process the data
 *    from the 'probe'; return a FALSE value otherwise.
 *
 *    Invoked by reader_type->want_probe_fn();
 */
static int readerWantProbe(skpc_probe_t *probe)
{
    /* This is what we expect: a directory based IPFIX listener */
    if ((skpcProbeGetPollDirectory(probe) != NULL)
        && (PROBE_ENUM_IPFIX == skpcProbeGetType(probe)))
    {
        return 1;
    }

    return 0;
}



/*
 *  status = readerSetup(&out_daemon_mode, probe_vector, options);
 *
 *    The reader_type should do any setup it requires prior to
 *    starting its flow_processor instances: Validate and/or store the
 *    reader_type's 'options'.  'probe_vector' is the list of probes
 *    that this reader_type said it would process
 *    (readerWantProbe).  'out_daemon_mode' should be set according
 *    to how this reader_type expects to run.
 *
 *    Invoked by reader_type->setup_fn();
 */
static int readerSetup(
    fp_daemon_mode_t   *is_daemon,
    const sk_vector_t  *probe_vec,
    reader_options_t   *options)
{
    /* this function should only be called if we actually have probes
     * to process */
    if (0 == skVectorGetCount(probe_vec)) {
        return 1;
    }

    ipfixdir_list = skDequeCreate();
    if (ipfixdir_list == NULL) {
        return 1;
    }

    polling_interval = options->stream_polldir.polling_interval;

    /* Is a daemon */
    *is_daemon = FP_DAEMON_ON;

    return 0;
}


/*
** Local Variables:
** mode:c
** indent-tabs-mode:nil
** c-basic-offset:4
** End:
*/
